"""
Development utilities
====================================
Tools for developers.
"""

import click
import sh
from commands.helpers import TableOutput as T
from os.path import isfile, isdir
from dotenv import load_dotenv

# declare the commands globally
git = sh.Command('git')

@click.command(context_settings={"ignore_unknown_options": True})
@click.argument('arguments', nargs=-1)
def xpgl(arguments):
  """
    XPGL is a wrapper around git log. It provides you a tabular output of 
    the most recent git commits
  """
  rows = []
  cnt = 1
  for line in git('--no-pager', 'log', '-n', '10', *arguments, '--pretty=%h#%d#%s#%cn#%cr'):
    cols = line.rstrip('\n').split('#')
    cols[1], cols[2], cols[3], cols[4] = cols[3], cols[4], cols[2], cols[1]
    cols.insert(0, str(cnt))
    rows.append('#'.join(cols))
    cnt += 1
  T.out(rows, headers=('N', 'Hash', 'Who', 'When', 'Message', 'Info'))

if __name__ == '__main__':
  xpgl()


"""
Dockertools
====================================
Tools to manage Docker.
"""

import os
import sys
import click
import sh
from commands.helpers import Executor as E, Message as M, TableOutput as T
from os.path import isfile, isdir
from dotenv import load_dotenv

# declare the commands globally
bash = sh.Command('bash')
docker = sh.Command('docker')


@click.command(context_settings={"ignore_unknown_options": True})
@click.option('--shell', '-s', is_flag=True, default=False, help='open a shell to the container, basically exec')
@click.option('--logs', '-l', is_flag=True, default=False, help='follows the logs, basically logs -f')
@click.option('--ignore-override', '-i', is_flag=True, default=False, help='ignores the docker-compose.override.yml')
@click.argument('docker_arguments', nargs=-1)
def dcc(docker_arguments, shell=False, logs=False, ignore_override=False):
    """
    DCC is a wrapper around docker-compose. If you are executing it inside a git
    repository it will set the following environment variables:
    - GIT_VERSION
    - GIT_BRANCH
    - GIT_LASTCOMMITDATE
    - GIT_COMMITHASH
    You can use those in the build process to pass the git version information to the application. 

    Additionally it handles the docker-compose files. You can set the environment variable DCC_ENV to
    whatever string is in between docker-compose.<????>.yml. DCC will then take docker-compose.yml as
    the base file, the next is the docker-compose.<????>.yml and if a docker-compose.override.yml exists
    it will take this as the last compose file in the -f ordering.
    """
    if not isfile('docker-compose.yml'):
        M.error('There is no docker-compose.yml file in this directory!')
        sys.exit(1)
    dcc_env = None
    env = {
        'GIT_VERSION': '?.?.?',
        'GIT_BRANCH': 'N/A',
        'GIT_LASTCOMMITDATE': 'N/A',
        'GIT_COMMITHASH': 'N/A'
    }
    if isfile('.env'):
        load_dotenv('.env')
        dcc_env = os.environ.get('DCC_ENV')
        M.debug(f"Running DCC environment {dcc_env}.")
    else:
        M.warn(f"No .env file found in current directory: {os.getcwd()}")
    if isdir('.git'):
        if E.success('git rev-parse --is-inside-work-tree'):
            env['GIT_VERSION'] = E.run('git describe --always')
            env['GIT_BRANCH'] = E.run('git rev-parse --abbrev-ref HEAD')
            env['GIT_LASTCOMMITDATE'] = E.run('git log -1 --format=%cI')
            env['GIT_COMMITHASH'] = E.run('git rev-parse HEAD')
        M.debug('Git repository detected.')
    else:
        M.warn("Not a git repository.")

    docker_files = [
        'docker-compose.yml'
    ]
    if dcc_env:
        docker_files.append(f"docker-compose.{dcc_env}.yml")
    if not ignore_override and isfile('docker-compose.override.yml'):
        docker_files.append('docker-compose.override.yml')
    if shell:
        docker_arguments = ['exec', *docker_arguments, '/bin/bash']
    elif logs:
        docker_arguments = ['logs', '-f', *docker_arguments]
    command = f"docker-compose -f {' -f '.join(docker_files)} {' '.join(docker_arguments)}"
    M.debug(f"Command: {command}")
    os.environ.update(env)
    try:
        bash('-c', command, _fg=True)
    except Exception as ex:
        M.error(f"Error executing command {command}: {ex}")


@click.command()
@click.argument('filter', required=False)
def dtls(filter: str = None):
    _filter = ''
    if filter:
        _filter = f"--filter name={filter}"
    command = "docker ps %s -a --format '{{.Names}}#{{.Status}}#{{.Image}}'" % _filter
    output = E.run(command)
    T.out(output, headers=('Name', 'Status', 'Image'))


@click.command()
@click.argument('filter', required=False)
def dtins(filter: str = None):
    rows = []
    args = ['ps', '-a', '--format', '{{.Names}}']
    if filter:
        args.append('--filter ')
        args.append(f"name={filter}")
    for container in docker(*args, _iter=True):
        container = container.rstrip('\n')
        insp = docker('inspect', '-f',
                      '{{ .State.Status }}#{{ .HostConfig.RestartPolicy.Name }}', container)
        rows.append(f"{container}#{insp}")
    T.out(rows, headers=('Name', 'Status', 'Restart'))


def __get_container_name(pattern):
    containers = E.run('docker ps -a --format "{{.Names}}"').split('\n')
    found_containers = []
    for container in containers:
        if pattern in container:
            found_containers.append(container)
    if len(found_containers) == 1:
        return found_containers[0]
    elif len(found_containers) == 0:
        M.error(f"No container found with name {pattern}.")
        raise
    else:
        M.error(
            f"More than one containers found with name {pattern}: {', '.join(found_containers)}")
        raise


@click.command()
@click.argument('pattern')
def dtail(pattern):
    try:
        docker('logs', '-f', __get_container_name(pattern), _fg=True)
    except:
        pass


@click.command()
@click.argument('pattern')
def dtsh(pattern):
    try:
        docker('exec', '-it', __get_container_name(pattern),
               '/bin/bash', _fg=True)
    except sh.ErrorReturnCode as ex:
        try:
            docker('exec', '-it', __get_container_name(pattern),
                   '/bin/sh', _fg=True)
        except:
            pass
    except:
        pass


@click.command()
@click.option('--force', '-f', is_flag=True, default=False, help='force removal, do not ask')
def dtclean(force):
    try:
        argss = [
            ['image', 'prune'],
            ['image', 'prune', '-a'],
            ['container', 'prune'],
            ['system', 'prune'],
            ['builder', 'prune', '-a']
        ]
        for args in argss:
            if force:
                args.append('-f')
            docker(*args, _fg=True)
    except:
        pass



"""
Helpers
====================================
Helper classes and functions
"""

import os
import shlex
from subprocess import Popen, run
from enum import Enum
from yachalk import chalk
from rich.console import Console
from rich.table import Table


class ReturnCode(Enum):
    OK = 0


class Executor():
    """
    Class to execute shell commands
    """
    @classmethod
    def success(cls, command: str, env: dict = None) -> bool:
        ret, _ = cls.__run(command, env=env)
        return ret == ReturnCode.OK.value

    @classmethod
    def run(cls, command: str, env: dict = None) -> str:
        _, output = cls.__run(command, env=env)
        return cls.__handle_output(output)

    @classmethod
    def __run(cls, command: str, env: dict = None) -> tuple[int, str]:
        """
        Runs a shell command with the default environment.
        If env is given it is _UPDATED_ to the default environment.
        """
        _env = os.environ.copy()
        if env:
            _env.update(env)
        p = run(shlex.split(command), capture_output=True, env=_env)
        return p.returncode, cls.__handle_output(p.stdout.decode())

    @classmethod
    def __handle_output(cls, output: str) -> str:
        # We remove the last \n
        return output.rstrip('\n')


class Message():
    """
    Class to output colored messages to the console
    """
    @classmethod
    def info(cls, message):
        print(chalk.green_bright.bold(message))

    @classmethod
    def warn(cls, message):
        print(chalk.yellow_bright.bold(message))

    @classmethod
    def error(cls, message):
        print(chalk.red_bright.bold(message))

    @classmethod
    def debug(cls, message):
        print(chalk.blue_bright.bold(message))


class TableOutput():
    """
    Class to output text in table format
    """
    console = Console()

    @classmethod
    def out(cls, data: str | list, sep: str = '#', headers: tuple[str] = None):
        table = Table(
            show_header=(headers is not None),
            show_lines=False,
            show_edge=False
        )
        if headers:
            for header in headers:
                table.add_column(header)

        if isinstance(data, str):
            data = data.split('\n')

        for line in data:
            table.add_row(*line.split(sep))

        cls.console.print(table)


####################
# OLD CODE
####################


# class Interactive():
#     """
#     Execute a command interactively with pseudo terminal
#     found at https://stackoverflow.com/questions/41542960/run-interactive-bash-with-popen-and-a-dedicated-tty-python
#     """

#     def __init__(self, command: str = '/bin/bash', env: dict = None):
#         self.command = command
#         self.env = os.environ.copy()
#         if env:
#             self.env.update(env)
#         self.process = None

#     def run(self):
#         old_tty = termios.tcgetattr(sys.stdin)
#         tty.setraw(sys.stdin.fileno())

#         master, slave = pty.openpty()

#         try:
#             self.process = Popen(
#                 shlex.split(self.command),
#                 preexec_fn=os.setsid,
#                 stdin=slave,
#                 stdout=slave,
#                 stderr=slave,
#                 env=self.env)

#             while True:
#                 r, _, _ = select.select([sys.stdin, master], [], [])
#                 if sys.stdin in r:
#                     d = os.read(sys.stdin.fileno(), 10240)
#                     os.write(master, d)
#                 elif master in r:
#                     o = os.read(master, 10240)
#                     if o:
#                         os.write(sys.stdout.fileno(), o)

#                 if self.process.poll() is not None:
#                     sys.stdout.flush()
#                     break
#                 else:
#                     sleep(0.1)

#         finally:
#             # restore tty settings back
#             termios.tcsetattr(sys.stdin, termios.TCSADRAIN, old_tty)



"""
SSH utils
====================================
Tools to manage SSH.
"""

import click
from sh import ssh


@click.command()
@click.option('--port', '-p', type=int, default=22, help='the port to connect to')
@click.option('--identity', '-i', help='name of the key (should be placed inside ~/.ssh/...)')
@click.option('--force-password-auth', '-f', is_flag=True, default=False,  help='force password authentication')
@click.option('--disable-host-key-checking', '-d', is_flag=True, default=False, help='do not check if the host key is correct [!! be careful !!]')
@click.option('--verbose', '-v', is_flag=True, default=False, help='display verbose output')
@click.argument('connection')
def sussh(connection, port=22, identity=None, force_password_auth=False, disable_host_key_checking=False, verbose=False):
    """Extended ssh command"""
    _options = ''
    if force_password_auth:
        _options += ' -o PreferredAuthentications=password -o PubkeyAuthentication=no'
    if disable_host_key_checking:
        _options += ' -o UserKnownHostsFile=/dev/null -o StrictHostKeyChecking=no'

    _identity = ''
    if identity:
        _identity = f"-i ~/.ssh/{identity}"

    command = f"ssh {_options} {_identity} -p {port} {connection}"

    if verbose:
        print('RUNNING:', command)
    ssh(command.split())


@click.command()
@click.argument('connection')
@click.argument('rport')
@click.argument('lport')
def surtun(connection, rport, lport):
    ssh('-4', '-N', '-T', '-R', f"{rport}:localhost:{lport}", connection)


@click.command()
@click.argument('connection')
@click.argument('rport')
@click.argument('lport')
def suftun(connection, rport, lport):
    ssh('-4', '-N', '-T', '-L', f"{lport}:localhost:{rport}", connection)



from setuptools import setup, find_packages

long_description = 'Power User Tools make your life so much easier.'

setup(
    name='power-user-tools',
    version='0.0.7',
    author='dameyerdave',
    author_email='dameyerdave@gmail.com',
    url='https://github.com/dameyerdave/power-user-tools',
    description='Power User Tools make your life so much easier',
    long_description=long_description,
    long_description_content_type="text/markdown",
    license='MIT',
    packages=find_packages(),
    entry_points={
            'console_scripts': [
                'sussh = commands.sshutils:sussh',
                'surtun = commands.sshutils:surtun',
                'dcc = commands.dockertools:dcc',
                'dtls = commands.dockertools:dtls',
                'dtail = commands.dockertools:dtail',
                'dtsh = commands.dockertools:dtsh',
                'dtclean = commands.dockertools:dtclean',
                'dtins = commands.dockertools:dtins',
                'xpgl = commands.devutils:xpgl',
            ]
    },
    scripts=[
        'shell/dtip',
        'shell/dtports',
        'shell/dtnet',
        'shell/dtimg',
        'shell/dtstart',
        'shell/dtstop',
        'shell/dtrestart',
        'shell/dtrm',
        'shell/xppr',
        'shell/xprandpw',
        'shell/xpgrmhistory',
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    keywords='power user tools ssh sshd reverse tunnel docker easy development',
    install_requires=[
        'click==8.1.3',
        'python-dotenv==0.21.0',
        'friendlylog==1.0.2',
        'yachalk==0.1.5',
        'rich==12.6.0',
        'sh==1.14.3',
    ],
    zip_safe=False
)
